/**
 * Written by Frits van Bommel, 2008.
 * Feel free to use this code however you want, but please keep my name in a
 * comment.
 * 
 * Note: This probably won't work on non-posix platforms (it requires glibc) so
 * I put a 'version(Posix):' at the top to keep it from failing to link on
 * e.g. Windows.
 */

module trace.stacktrace;

version(Posix):     // Actually requires glibc, not posix.

import trace.demangle;

import tango.text.Util;

import tango.stdc.stdlib;
import tango.stdc.stringz;

private {
    // From glibc's execinfo.h
    extern(C) {
        // Returns number of return addresses stored to arr (up to size)
        int backtrace (void** arr, int size);
        // Returns function names of backtrace in newly-malloc()ed memory
        char** backtrace_symbols (void** arr, int size);
        // void backtrace_symbols_fd (void** arr, int size, int fd)
    }
    
    // from tango/lib/compiler/*/genobj.d :
    alias Exception.TraceInfo function(void* ptr = null) TraceHandler;
    extern (C) void  rt_setTraceHandler(TraceHandler h);

}

private Exception.TraceInfo traceHandler(void* ptr = null) {
    return new Backtrace;
}

// Install backtrace handler
static this() {
    rt_setTraceHandler(&traceHandler);
}

/**
 * Class that provides backtrace information from the point where its
 * constructor is called.
 */
class Backtrace : Exception.TraceInfo {
    /**
     * Initial lines containing this string will be filtered out.
     * Make sure you change this if you change the module name!
     */
    private static const char[] filterStr = "_D5trace10stacktrace";
    
    /// The maximum number of frames we'll try to get if the trace is long.
    public static const size_t maxFrames = 4096;
    
    static assert(maxFrames <= cast(size_t)int.max);
    
    /// The return addresses of the stackframes of the trace.
    const void*[] addresses;
    /// Strings representing the stackframes of the trace.
    private char[][] trace_;
    
    /// Default constructor. Gets a backtrace and stores it for later use.
    this() {
        addresses = GetAddresses();
        trace_ = null;
    }
    
    /// Returns strings representing the frames of the trace.
    char[][] trace() {
        if (trace_ is null && addresses.length > 0)
            trace_ = AddressesToNames(addresses);
        return trace_;
    }
    
    /// Iterates over the frames in the backtrace.
    int opApply(int delegate(inout char[]) dg) {
        int ret = 0;
        foreach (frame; trace) {
            ret = dg( frame );
            if (ret)
                break;
        }
        return ret;
    }
    
    /// Returns the backtrace as a single string.
    char[] toString() {
        return trace().join("\n");
    }
    
    /// Returns a backtrace for its invocation as an array of lines.
    private static void*[] GetAddresses() {
        // Get return addresses
        void*[maxFrames] buf;
        auto filled = backtrace(buf.ptr, buf.length);
        return buf[0 .. filled].dup;
    }
    
    /// Turns a raw list of return addresses into a human-readable stack trace.
    private static char[][] AddressesToNames(void*[] addrs) {
        if (addrs.length > cast(size_t)int.max)
            addrs.length = int.max;
        
        // Match them to names.
        auto cnames = backtrace_symbols(addrs.ptr, addrs.length);
        scope(exit)
            free(cnames);
        
        auto start = cnames;
        auto end = cnames + addrs.length;
        
        // Filter out initial frames from this module
        while (start < end && fromStringz(*start).containsPattern(filterStr))
            ++start;
        
        // Convert to char[]s and process
        auto trace = new char[][](end - start);
        foreach (i, cline; start[0 .. end - start]) {
            trace[i] = process(fromStringz(cline));
        }
        
        return trace;
    }
    
    /**
     * Demangles any D function names in the passed backtrace line.
     * Always returns a fresh gc-allocated string, since the memory returned
     * by backtrace_symbols is one big malloced block that gets freed
     * by Get()
     */
    private static char[] process(char[] line) {
        // Lines seem to have format <programname>(<fn>+<offset>) [addr]
        // where the parenthesized part and the '+<offset>' part within it are
        // both optional.
        
        // Find the opening parenthesis
        auto begin = line.locate('(');
        
        if (begin < line.length) {
            // A function name is present, see if we need to demangle it.
            
            ++begin; // now index of first char of function name
            
            // Find the end of the function name.
            // (Either the ')' or the last '+' before it terminates it)
            auto end = line.locatePrior(')');
            end = line[begin .. end].locatePrior('+') + begin;
            
            auto funcName = line[begin .. end];
            if (end - begin > 2 && funcName[0 .. 2] == "_D") {
                // This looks like a mangled D name, demangle and replace:
                return line[0 .. begin] ~ demangle(funcName) ~ line[end .. $];
            }
        }
        // No name to demangle, return a copy of the original.
        return line.dup;
    }
}
